#ifdef HAVE_CONFIG_H
    #include "config.h"
#endif
#include "carnac_nussinov_stems.h"
#include "display.h"
#include "opts.h"


#include <stdlib.h>
#include <assert.h>



void
nussi_computeTable_Cell(int *table, int *graph, const int i, const metastem_t_list ostems, const stem_open_index_t *open_at, const stem_close_index_t *close_at) {
   int w, x;
   int subA;



   subA = 1+open_at[i-1].open_sub;
   
   /*
     INFO_ "*** TABLE CELL COMPUTING SUB [%i..%i] [%i..%i]\n", subA, (1+open_atA[i-1].close_sub), subB,(1+open_atB[j-1].close_sub)  _INFO;
   */


   table[subA-1] = 0;
   
   
   for(x = subA; x <= 1+open_at[i-1].close_sub; x++) {
      
      /*
	INFO_ "  tc[%i][%i] table[%2i][%2i] = MIN(", i,j,x, y _INFO;
      */
      
      table[x] = table[x-1];
      
      
      /*
	INFO_ "%5i, %5i", table[x-1][y], table[x][y-1] _INFO;
      */
      
      if((1+close_at[x-1].open_index) >= i) {
	 
	 w = graph[x];
	 
	 /*
	   INFO_ ", g = %5i", w _INFO;
	 */
	 
	 if((open_at[close_at[x-1].open_index].close_prev >= 0) &&
	    ((1+open_at[close_at[x-1].open_index].close_prev) >= subA)) {
	    
	    w += table[1+open_at[close_at[x-1].open_index].close_prev];
	    /*
	      INFO_ ", g' = %5i", w _INFO;
	    */
	 }
	 
	 table[x] = MIN(table[x], w);
      }
      
      /*
	INFO_ ") = %5i\n", table[x][y] _INFO;
      */
      
   }
}



void
nussi_add_edges(const metastem_t_list ostems, const metastem_list stems, const stem_open_index_t *open_at, const stem_close_index_t *close_at, const int i, const int x, int *table, int *graph) {

   short sap, san;
   short sapi, sani;



   /*INFO_ "ADD EDGE %i, %i, %i, %i\n", i, x, j, y _INFO;*/
   
   for(sap = 0; sap < STACK_SIZE(open_at[i-1].open_for); sap++) {
      sapi = STACK_VALUE_AT(open_at[i-1].open_for, sap);
      for(san = 0; san < STACK_SIZE(close_at[x-1].close_for); san++) {
	 sani = STACK_VALUE_AT(close_at[x-1].close_for, san);
	 if(open_at[sapi].stem_id == close_at[sani].stem_id) {
	    
	    /*INFO_ "   CLOSING INTERVAL [%i .. %i] for stem_id = %i\n", i-1,x-1, close_at[sani].stem_id _INFO;*/
	    
	    
	    graph[1+sani] = table[x] + ostems->metastems[open_at[sapi].stem_id].energy;
	    
			/*INFO_ "    graph[%3i][%3i] = %5i with table[%3i][%3i] = %5i\n", 1+sani, 1+sbni, table[1+sani][1+sbni].stem_energy, x, y, table[x][y].energy _INFO;*/

	 }
      }
   }
}




void
nussi_build_alignment_graph(const metastem_t_list ostems, const metastem_list stems, const stem_open_index_t *open_at, const stem_close_index_t *close_at, int *table, int *graph) {

   int i,x;
   int olsA, clsA;



   for (i = stems.nb_metastems; i > 0; i--) {
      if(open_at[i-1].is_hairpin) {
	 
	 graph[1+open_at[i-1].close_index] = ostems->metastems[open_at[i-1].stem_id].energy;
	 
	 /*
	   INFO_ "    HP graph[%3i][%3i] = %5i\n", 1+open_atA[i-1].close_index, 1+open_atB[j-1].close_index, graph[1+open_atA[i-1].close_index][1+open_atB[j-1].close_index] _INFO;
	 */
	 
      }
   }

   
   olsA = -1;
   clsA = -1;
   

   /*****************************/ 
   /*    weight of the edges    */
   /*****************************/
   for (i = stems.nb_metastems; i > 0; i--) {
      
      if(open_at[i-1].open_for) {
	 
	 /*INFO_ "********** I = %i, J = %i\n", i, j _INFO;*/
	 
	 
	 if((olsA != 1+open_at[i-1].open_sub) ||
	    (clsA != 1+open_at[i-1].close_sub)) {
	    
	    olsA = 1+open_at[i-1].open_sub;
	    clsA = 1+open_at[i-1].close_sub;
	    
	    nussi_computeTable_Cell(table, graph, i, ostems, open_at, close_at);
	 }
	 
	 
	 for(x = 1+open_at[i-1].open_sub; x <= (1+open_at[i-1].close_sub); x++) {
	    if(close_at[x-1].close_for) {
	       
	       /*	 INFO_ "TRYING (i=%i, j=%i, x=%i, y=%i)\n", i, j, x, y _INFO;*/
	       
	       nussi_add_edges(ostems, stems, open_at, close_at, i,x, table, graph);
	    }
	 }
      }
   }
}






void
nussi_computeTable(const metastem_list stems, const stem_open_index_t *open_at, const stem_close_index_t *close_at, int *table, int *graph) {
   int w, x;


   table[0] = 0;
   

   for(x = 1; x <= stems.nb_metastems; x++) {

	 /*INFO_ "table[%2i][%2i] = MIN(", x, y _INFO;*/

	 table[x] = table[x-1];
	 
	 /*	 INFO_ "%5i, %5i", table[x-1][y], table[x][y-1] _INFO;*/
	 
	 w = graph[x];
	 
	 /*	 INFO_ ", g = %5i", w _INFO;*/
	 
	 if(open_at[close_at[x-1].open_index].close_prev >= 0) {

	    /*	    INFO_ ", g' = %5i", table[1+open_atA[close_atA[x-1].open_index].close_prev_index][1+open_atB[close_atB[y-1].open_index].close_prev_index] _INFO;*/

	    w += table[1+open_at[close_at[x-1].open_index].close_prev];
	 }

	 table[x] = MIN(table[x], w);


	 /*	 INFO_ ") = %5i\n", table[x][y] _INFO;*/
   }

   /*INFO_ "ENERGY GLOBALE : %i\n", table[stemsA.nb_stems][stemsB.nb_stems] _INFO;*/
}








void
nussi_build_optimal_folding(const int i, const int k, const metastem_t_list ostems, const stem_open_index_t *open_at, const stem_close_index_t *close_at, int *table, stack *p, metastem_t_list_t *res) {

   int x;

   x = k;

   /*INFO_ "  OPTIMAL FOLDING : i = %i, k = %i, j = %i, l = %i\n",i,k,j,l _INFO;*/
   

   while(x>=i) {
      if(table[x] == table[x-1]) {
	 /*INFO_ "  SKIP x = %i, y = %i (%i) TO x = %i, y = %i (%i)\n", x, y, table[x][y].energy,x-1,y, table[x-1][y].energy _INFO;*/
	 x--;
      } else {

	 /*INFO_ "  TAKE x = %i, y = %i (%i) JUMP TO x = %i, y = %i (%i, stem+under = %i)\n", x, y, table[x][y].energy, 1+open_atA[close_atA[x-1].open_index].close_prev, 1+open_atB[close_atB[y-1].open_index].close_prev, table[1+open_atA[close_atA[x-1].open_index].close_prev][1+open_atB[close_atB[y-1].open_index].close_prev].energy, table[x][y].stem_energy _INFO;*/
	 
	 if(!open_at[close_at[x-1].open_index].is_hairpin) {
	    /* should look for subopt */
	    push(p, x, x);
	    /*INFO_ "  PUSH %i, %i\n", x, y _INFO;*/
	 }

	 if(res->nb_metastems == 0) {
	    NEW(res->metastems, metastem_t, 1);
	 } else {
	    RENEW(res->metastems, metastem_t, (res->nb_metastems+1));
	 }
	 
	 res->metastems[res->nb_metastems] = ostems->metastems[close_at[x-1].stem_id];
	 /*res->energy += ostems->stems[close_at[x-1].stem_id].energy;*/

	 res->nb_metastems++;
	 
	 x = 1+open_at[close_at[x-1].open_index].close_prev;
      }
   }
}







void
nussi_trace_back(const metastem_t_list ostems, const metastem_list stems, const stem_open_index_t *open_at, const stem_close_index_t *close_at, int *table, int *graph, metastem_t_list_t *res) {
   stack *p = NULL;
   int x,y;


   NEW(p, stack, 1);

   init_stack(p, stems.nb_metastems);
   
   /*INFO_ "TRACE BACK\n" _INFO;*/

   nussi_computeTable(stems, open_at, close_at, table, graph);
   
   x = stems.nb_metastems;
   nussi_build_optimal_folding(1, x, ostems, open_at, close_at, table, p, res);


   while(!is_empty_stack(p)) {
      pop(p, &x,&y);
      /*INFO_ "POP %i, %i\n", x, y _INFO;*/
      
      nussi_computeTable_Cell(table, graph, 1+open_at[close_at[x-1].open_index].open_next, ostems, open_at, close_at);

      nussi_build_optimal_folding(1+open_at[open_at[close_at[x-1].open_index].open_next].open_sub, 1+close_at[x-1].close_prev, ostems, open_at, close_at, table, p, res);
      
   }


   free_stack(p);
}












metastem_t_list_t*
nussinov_stems(const metastem_t_list ostems, const metastem_list stems, const stem_open_index_t *open_at, const stem_close_index_t *close_at) {
   metastem_t_list_t *res = NULL;
   
   int i;
   int *table = NULL;
   int *graph = NULL;

   
   NEW(table, int, (stems.nb_metastems+1));
   NEW(graph, int, (stems.nb_metastems+1));
   for(i = 0; i <= stems.nb_metastems; i++) {
      table[i] = 0;
      graph[i] = 0;
   }
   
   nussi_build_alignment_graph(ostems, stems, open_at, close_at, table, graph);
   
   NEW(res, metastem_t_list_t, 1);
   res->nb_metastems = 0;
   res->metastems = NULL;
   nussi_trace_back(ostems, stems, open_at, close_at, table, graph, res);
   
   
   DESTROY(table);
   DESTROY(graph);

   return res;
}










