#ifdef HAVE_CONFIG_H
    #include "config.h"
#endif
#include "carnac_sankoff_stems.h"

#include "opts.h"
#include "display.h"


#include <stdlib.h>
#include <assert.h>




void
prepare_sankoff_stems(const metastem_t_list ostems, const metastem_list stems, stem_open_index_t **oindex, stem_close_index_t **cindex) {
   short i, j, k;
   metastem_list sbegin, sclose;

   stem_index_t *res = NULL;
   stem_open_index_t *reso = NULL;
   stem_close_index_t *resc = NULL;
   
   NEW(res, stem_index_t, (2*stems.nb_metastems));
   NEW((*oindex), stem_open_index_t, (stems.nb_metastems));
   reso = (*oindex);
   NEW((*cindex), stem_close_index_t, (stems.nb_metastems));
   resc = (*cindex);

   NEW(sbegin.metastem_id, int, stems.nb_metastems);
   NEW(sclose.metastem_id, int, stems.nb_metastems);


   for(i = 0; i < stems.nb_metastems; i++) {
      sbegin.metastem_id[i] = sclose.metastem_id[i] = stems.metastem_id[i];
   }
   
   sbegin.nb_metastems = sclose.nb_metastems = stems.nb_metastems;

   sort_metastems_id_by_begin_then_length(&(sbegin), ostems);
   sort_metastems_id_by_begin_closure_then_length(&(sclose), ostems);



   i = 0;
   j = 0;
   
   while((i < stems.nb_metastems) && (j < stems.nb_metastems)) {

      if(begin_start_leq_end_start(ostems->metastems[sbegin.metastem_id[i]], ostems->metastems[sclose.metastem_id[j]], 0)) {

	 reso[i].stem_id = sbegin.metastem_id[i];
	 reso[i].close_index = -1;
	 reso[i].open_next = -1;
	 reso[i].close_next = -1;
	 reso[i].close_prev = -1;
	 reso[i].open_sub = -1;
	 reso[i].close_sub = -1;
	 reso[i].is_hairpin = 1;
	 reso[i].open_for = NULL;
	 
	 res[i+j].stem_id = sbegin.metastem_id[i];
	 res[i+j].type = STEM_OPENING;
	 res[i+j].other_index = -1; /* unknown at the moment */
	 res[i+j].offset = i;

         i++;
      } else {
	 resc[j].stem_id = sclose.metastem_id[j];
	 resc[j].close_prev = -1;
	 resc[j].close_for = NULL;

	 res[i+j].stem_id = sclose.metastem_id[j];
	 res[i+j].type = STEM_CLOSING;
	 res[i+j].offset = j;

	 k = i+j-1;
	 while(res[k].stem_id != res[i+j].stem_id) { k--; } /* looking for opening */
	 res[i+j].other_index = k;
	 res[k].other_index = i+j; /* updating opening */

	 resc[j].open_index = res[k].offset;
	 reso[res[k].offset].close_index = j;

         j++;
      }
   }

   while(i < stems.nb_metastems) {
      reso[i].stem_id = sbegin.metastem_id[i];
      reso[i].close_index = -1;
      reso[i].open_next = -1;
      reso[i].close_next = -1;
      reso[i].close_prev = -1;
      reso[i].open_sub = -1;
      reso[i].close_sub = -1;
      reso[i].is_hairpin = 1;
      reso[i].open_for = NULL;
      
      res[i+j].stem_id = sbegin.metastem_id[i];
      res[i+j].type = STEM_OPENING;
      res[i+j].other_index = -1; /* unknown at the moment */
      res[i+j].offset = i;
      
      i++;
   }
   
   while(j < stems.nb_metastems) {
      resc[j].stem_id = sclose.metastem_id[j];
      resc[j].close_prev = -1;
      resc[j].close_for = NULL;
      
      res[i+j].stem_id = sclose.metastem_id[j];
      res[i+j].type = STEM_CLOSING;
      res[i+j].offset = j;
      
      k = i+j-1;
      while(res[k].stem_id != res[i+j].stem_id) { k--; } /* looking for opening */
      res[i+j].other_index = k;
      res[k].other_index = i+j; /* updating opening */
      
      resc[j].open_index = res[k].offset;
      reso[res[k].offset].close_index = j;
      
      j++;
   }


   DESTROY(sbegin.metastem_id);
   DESTROY(sclose.metastem_id);


   /* open_next */
   for(i = 0; i < 2*stems.nb_metastems; i++) {

      switch(res[i].type) {
      case STEM_OPENING:

	 /* skipping everything overlapping */
	 j = i+1;
	 while((j < res[i].other_index)
	       && (reso[res[i].offset].open_next < 0)) {
	    if((res[j].type == STEM_OPENING) && !begin_start_leq_begin_stop(ostems->metastems[res[j].stem_id], ostems->metastems[res[i].stem_id], -CARNAC_TOLERATE_OVERLAP)) {
	       reso[res[i].offset].open_next = res[j].offset;
	    }

	    j++;
	 }
	 /**/

	 break;
      case STEM_CLOSING:
	 break;
      }
   }
   /**/




   /* close_next */
   for(i = 0; i < 2*stems.nb_metastems; i++) {

      switch(res[i].type) {
      case STEM_OPENING:

	 /* skipping everything overlapping */
	 j = i+1;
	 while((j < res[i].other_index)
	       && (reso[res[i].offset].close_next < 0)) {
	    if((res[j].type == STEM_CLOSING) && !end_start_leq_begin_stop(ostems->metastems[res[j].stem_id], ostems->metastems[res[i].stem_id], -CARNAC_TOLERATE_OVERLAP)) {
	       reso[res[i].offset].close_next = res[j].offset;
	    }

	    j++;
	 }
	 /**/

	 break;
      case STEM_CLOSING:
	 break;
      }
   }
   /**/



   /* close_prev */
   for(i = 2*stems.nb_metastems - 1; i >= 0; i--) {
      
      switch(res[i].type) {
      case STEM_OPENING:
	 break;
      case STEM_CLOSING:
	 
	 j = i+1;
	 k = 1;
	 while(k
	       && (j < 2*stems.nb_metastems)) {
	    switch(res[j].type) {
	    case STEM_OPENING:
	       if(reso[res[j].offset].close_prev < 0) {

		  if(!begin_start_leq_end_stop(ostems->metastems[res[j].stem_id], ostems->metastems[res[i].stem_id], -CARNAC_TOLERATE_OVERLAP)) {
		     reso[res[j].offset].close_prev = res[i].offset;
		  }

	       } else {
		  k = 0;
	       }       
	       break;
	    case STEM_CLOSING:
	       if(resc[res[j].offset].close_prev < 0) {

		  if(!end_start_leq_end_stop(ostems->metastems[res[j].stem_id], ostems->metastems[res[i].stem_id], -CARNAC_TOLERATE_OVERLAP)) {
		     resc[res[j].offset].close_prev = res[i].offset;
		  }

	       } else {
		  k = 0;
	       }
	       break;
	    }

	    j++;
	 }
      }
   }
   /**/





   /* detecting hairpins */
   for(i = 0; i < 2*stems.nb_metastems; i++) {
      switch(res[i].type) {
      case STEM_OPENING:

	 if((reso[res[i].offset].open_next >= 0)
	    && (reso[res[i].offset].close_next >= 0)
	    && (resc[reso[res[i].offset].close_index].close_prev >= 0)) {
	    
	    for(j = reso[res[i].offset].close_next; j <= resc[reso[res[i].offset].close_index].close_prev; j++) {
	       
	       if(resc[j].open_index >= reso[res[i].offset].open_next) {
		  reso[res[i].offset].is_hairpin = 0;
	       }
	    }
	 }
	 break;
      case STEM_CLOSING:
	 break;
      }      
   }
   /**/









   /* open_for and close_for */
   for(i = 0; i < 2*stems.nb_metastems; i++) {
      switch(res[i].type) {
      case STEM_OPENING:
	 if((reso[res[i].offset].open_next >= 0) && !(reso[res[i].offset].is_hairpin)) {
	    if(reso[reso[res[i].offset].open_next].open_for) {
	       if(STACK_SIZE(reso[reso[res[i].offset].open_next].open_for) == reso[reso[res[i].offset].open_next].open_for->max_size) {
		  reso[reso[res[i].offset].open_next].open_for->max_size += 5;
		  RENEW(reso[reso[res[i].offset].open_next].open_for->elements, short, reso[reso[res[i].offset].open_next].open_for->max_size);
	       }
	    } else {
	       NEW_SHORT_STACK(reso[reso[res[i].offset].open_next].open_for,5);
	    }

	    STACK_PUSH(reso[reso[res[i].offset].open_next].open_for, res[i].offset);
	 }
	 break;
      case STEM_CLOSING:
	 if((resc[res[i].offset].close_prev >= 0) && !(reso[res[res[i].other_index].offset].is_hairpin)) {
	    if(resc[resc[res[i].offset].close_prev].close_for) {
	       if(STACK_SIZE(resc[resc[res[i].offset].close_prev].close_for) == resc[resc[res[i].offset].close_prev].close_for->max_size) {
		  resc[resc[res[i].offset].close_prev].close_for->max_size += 5;
		  RENEW(resc[resc[res[i].offset].close_prev].close_for->elements, short, resc[resc[res[i].offset].close_prev].close_for->max_size);
	       }
	    } else {
	       NEW_SHORT_STACK(resc[resc[res[i].offset].close_prev].close_for,5);
	    }

	    STACK_PUSH(resc[resc[res[i].offset].close_prev].close_for, res[i].offset);
	 }
	 break;
      }
   }
   /**/



   /* open_sub & close_sub */
   for(i = 0; i < 2*stems.nb_metastems; i++) {
      switch(res[i].type) {
      case STEM_OPENING:
	 if(reso[res[i].offset].open_for) {
	    for(j = 0; j < STACK_SIZE(reso[res[i].offset].open_for); j++) {
	       k = STACK_VALUE_AT(reso[res[i].offset].open_for, j);

	       reso[res[i].offset].close_sub = MAX(reso[res[i].offset].close_sub, resc[reso[k].close_index].close_prev);

	       if(reso[res[i].offset].open_sub < 0) {
		  if(reso[res[i].offset].close_next < 0) {
		     reso[res[i].offset].open_sub = res[res[i].other_index].offset;
		  } else {
		     reso[res[i].offset].open_sub = reso[res[i].offset].close_next;
		  }
	       }

	       if(reso[k].close_next >= 0) {
		  reso[res[i].offset].open_sub = MIN(reso[res[i].offset].open_sub, reso[k].close_next);
	       }
	    }
	 }
	 break;
      case STEM_CLOSING:
	 break;
      }      
   }



   for(i = 0; i < 2*stems.nb_metastems; i++) {
      switch(res[i].type) {
      case STEM_OPENING:
	 if(reso[res[i].offset].open_for) {
	    if((reso[res[i].offset].open_sub < 0) || (reso[res[i].offset].close_sub < 0)) {
	       DESTROY_STACK(reso[res[i].offset].open_for);
	    }
	 }
	 break;
      case STEM_CLOSING:
	 break;
      }
   }


   DESTROY(res);

}









void
computeTable_Cell(int **table, int **graph, const int i, const metastem_t_list ostemsA, const stem_open_index_t *open_atA, const stem_close_index_t *close_atA, const int j, const metastem_t_list ostemsB, const stem_open_index_t *open_atB, const stem_close_index_t *close_atB) {
   int w, x, y;
   int subA, subB;






   subA = 1+open_atA[i-1].open_sub;
   subB = 1+open_atB[j-1].open_sub;
   
   /*
     INFO_ "*** TABLE CELL COMPUTING SUB [%i..%i] [%i..%i]\n", subA, (1+open_atA[i-1].close_sub), subB,(1+open_atB[j-1].close_sub)  _INFO;
   */


   table[subA-1][subB-1] = 0;
   
   for(x = subA; x <= (1+open_atA[i-1].close_sub); x++) {
      table[x][subB-1] = 0;
   }
   
   for(y = subB; y <= (1+open_atB[j-1].close_sub); y++) {
      table[subA-1][y] = 0;
   }
   

   for(x = subA; x <= 1+open_atA[i-1].close_sub; x++) {
      for(y = subB; y <= 1+open_atB[j-1].close_sub; y++) {

	 /*
	   INFO_ "  tc[%i][%i] table[%2i][%2i] = MIN(", i,j,x, y _INFO;
	 */

	 table[x][y] = MIN(table[x-1][y], table[x][y-1]);


	 /*
	   INFO_ "%5i, %5i", table[x-1][y], table[x][y-1] _INFO;
	 */
	 
	 if(((1+close_atA[x-1].open_index) >= i) && ((1+close_atB[y-1].open_index) >= j)) {

	    w = graph[x][y];
	    
	    /*
	      INFO_ ", g = %5i", w _INFO;
	    */
	    
	    if((open_atA[close_atA[x-1].open_index].close_prev >= 0) &&
	       (open_atB[close_atB[y-1].open_index].close_prev >= 0) &&
	       ((1+open_atA[close_atA[x-1].open_index].close_prev) >= subA) &&
	       ((1+open_atB[close_atB[y-1].open_index].close_prev) >= subB)
	       ) {
	       
	       w += table[1+open_atA[close_atA[x-1].open_index].close_prev][1+open_atB[close_atB[y-1].open_index].close_prev];
	       /*
		 INFO_ ", g' = %5i", w _INFO;
	       */
	    }
	    
	    table[x][y] = MIN(table[x][y], w);
	 }
	 
	 /*
	   INFO_ ") = %5i\n", table[x][y] _INFO;
	 */

      }
   }
}



void
add_edges(const metasequence metaA, const metastem_t_list ostemsA, const metastem_list stemsA, const stem_open_index_t *open_atA, const stem_close_index_t *close_atA, const metasequence metaB, const metastem_t_list ostemsB, const metastem_list stemsB, const stem_open_index_t *open_atB, const stem_close_index_t *close_atB, const int i, const int j, const int x, const int y, int **table, int **graph, cofoldable_t **compatibles, int *olsA, int *olsB, int *clsA, int *clsB) {

   short sap, san, sbp, sbn;
   short sapi, sani, sbpi, sbni;


   /*INFO_ "ADD EDGE %i, %i, %i, %i\n", i, x, j, y _INFO;*/
   
      for(sap = 0; sap < STACK_SIZE(open_atA[i-1].open_for); sap++) {
	 sapi = STACK_VALUE_AT(open_atA[i-1].open_for, sap);

	    for(san = 0; san < STACK_SIZE(close_atA[x-1].close_for); san++) {
	       sani = STACK_VALUE_AT(close_atA[x-1].close_for, san);
	       if(open_atA[sapi].stem_id == close_atA[sani].stem_id) {
	    
		  /*INFO_ "   CLOSING INTERVAL [%i .. %i] for stem_id = %i\n", i-1,x-1, close_atA[sani].stem_id _INFO;*/
	    
		     for(sbp = 0; sbp < STACK_SIZE(open_atB[j-1].open_for); sbp++) {
			sbpi = STACK_VALUE_AT(open_atB[j-1].open_for, sbp);

			   for(sbn = 0; sbn < STACK_SIZE(close_atB[y-1].close_for); sbn++) {
			      sbni = STACK_VALUE_AT(close_atB[y-1].close_for, sbn);
			      if(open_atB[sbpi].stem_id == close_atB[sbni].stem_id) {
		     
				 /*INFO_ "       CLOSING INTERVAL [%i .. %i] for stem_id = %i\n", j-1,y-1, close_atB[sbni].stem_id  _INFO;*/
				 if(compatibles[open_atA[sapi].stem_id][open_atB[sbpi].stem_id].compatible
				    && compatibles[open_atA[sapi].stem_id][open_atB[sbpi].stem_id].do_covariate) {
				    /*INFO_ "COFOLDABLE %3i and %3i\n", open_atA[sapi].stem_id, open_atB[sbpi].stem_id _INFO;*/

				    
				    if((*olsA != 1+open_atA[i-1].open_sub) ||
				       (*olsB != 1+open_atB[j-1].open_sub) ||
				       (*clsA != 1+open_atA[i-1].close_sub) ||
				       (*clsB != 1+open_atB[j-1].close_sub)) {
				       
				       *olsA = 1+open_atA[i-1].open_sub;
				       *olsB = 1+open_atB[j-1].open_sub;
				       *clsA = 1+open_atA[i-1].close_sub;
				       *clsB = 1+open_atB[j-1].close_sub;
				       
				       computeTable_Cell(table, graph, i, ostemsA, open_atA, close_atA, j, ostemsB, open_atB, close_atB);
				    }

			
			
				    graph[1+sani][1+sbni] = table[x][y] + ostemsA->metastems[open_atA[sapi].stem_id].energy + ostemsB->metastems[open_atB[sbpi].stem_id].energy;
			
#if defined(CARNAC_NON_WATSON_CRICK) && defined(CARNAC_CONSERVED_NON_CANONICAL_BONUS)
				    if(compatibles[open_atA[sapi].stem_id][open_atB[sbpi].stem_id].have_non_canonical_conserved) {
				       graph[1+sani][1+sbni] += CARNAC_CONSERVED_NON_CANONICAL_BONUS;
				    }
#endif     
			
				    /*INFO_ "    graph[%3i][%3i] = %5i with table[%3i][%3i] = %5i\n", 1+sani, 1+sbni, table[1+sani][1+sbni].stem_energy, x, y, table[x][y].energy _INFO;*/
			
				 }
			      }
			   }
		     }

	       }
	    }

      
   }
}




void
build_alignment_graph(const metasequence metaA, const metastem_t_list ostemsA, const metastem_list stemsA, const stem_open_index_t *open_atA, const stem_close_index_t *close_atA, const metasequence metaB, const metastem_t_list ostemsB, const metastem_list stemsB, const stem_open_index_t *open_atB, const stem_close_index_t *close_atB, int **table, int **graph, cofoldable_t **compatibles) {

   int i,j,x,y;
   int olsA, clsA, olsB, clsB;



      
   olsA = -1;
   olsB = -1;
   clsA = -1;
   clsB = -1;
   


   if(OPTS_hairpin_mode) {
      for (i = stemsA.nb_metastems; i > 0; i--) {
	 for (j = stemsB.nb_metastems; j > 0; j--) {
	    if((compatibles[open_atA[i-1].stem_id][open_atB[j-1].stem_id].compatible)
	       && (compatibles[open_atA[i-1].stem_id][open_atB[j-1].stem_id].do_covariate)) {
	       
	       graph[1+open_atA[i-1].close_index][1+open_atB[j-1].close_index] = ostemsA->metastems[open_atA[i-1].stem_id].energy + ostemsB->metastems[open_atB[j-1].stem_id].energy;

	       if(!(open_atA[i-1].is_hairpin || open_atB[j-1].is_hairpin)) {
		  
		  if((olsA != (1+open_atA[open_atA[i-1].open_next].open_sub)) ||
		     (olsB != (1+open_atB[open_atB[j-1].open_next].open_sub)) ||
		     (clsA != (1+open_atA[open_atA[i-1].open_next].close_sub)) ||
		     (clsB != (1+open_atB[open_atB[j-1].open_next].close_sub))) {
		     
		     

		     olsA = 1+open_atA[open_atA[i-1].open_next].open_sub;
		     olsB = 1+open_atB[open_atB[j-1].open_next].open_sub;
		     clsA = 1+open_atA[open_atA[i-1].open_next].close_sub;
		     clsB = 1+open_atB[open_atB[j-1].open_next].close_sub;
		     
		     computeTable_Cell(table, graph, 1+open_atA[i-1].open_next, ostemsA, open_atA, close_atA, 1+open_atB[j-1].open_next, ostemsB, open_atB, close_atB);
		  }
		  
		  graph[1+open_atA[i-1].close_index][1+open_atB[j-1].close_index] += table[1+close_atA[open_atA[i-1].close_index].close_prev][1+close_atB[open_atB[j-1].close_index].close_prev];
		  
	       }
	       
	       
#if defined(CARNAC_NON_WATSON_CRICK) && defined(CARNAC_CONSERVED_NON_CANONICAL_BONUS)
	       if(compatibles[open_atA[i-1].stem_id][open_atB[j-1].stem_id].have_non_canonical_conserved) {
		  graph[1+open_atA[i-1].close_index][1+open_atB[j-1].close_index] += CARNAC_CONSERVED_NON_CANONICAL_BONUS;
	       }
#endif     
	       
	       
	       
	       /* INFO_ "    HP graph[%3i][%3i] = %5i\n", 1+open_atA[i-1].close_index, 1+open_atB[j-1].close_index, graph[1+open_atA[i-1].close_index][1+open_atB[j-1].close_index] _INFO;*/
	       
	       
	    }
	 }
      }





   } else {





      for (i = stemsA.nb_metastems; i > 0; i--) {
	 for (j = stemsB.nb_metastems; j > 0; j--) {
	    if((open_atA[i-1].is_hairpin || open_atB[j-1].is_hairpin)
	       && (compatibles[open_atA[i-1].stem_id][open_atB[j-1].stem_id].compatible)
	       && (compatibles[open_atA[i-1].stem_id][open_atB[j-1].stem_id].do_covariate)) {
	       
	       graph[1+open_atA[i-1].close_index][1+open_atB[j-1].close_index] = ostemsA->metastems[open_atA[i-1].stem_id].energy + ostemsB->metastems[open_atB[j-1].stem_id].energy;
	       
#if defined(CARNAC_NON_WATSON_CRICK) && defined(CARNAC_CONSERVED_NON_CANONICAL_BONUS)
	       if(compatibles[open_atA[i-1].stem_id][open_atB[j-1].stem_id].have_non_canonical_conserved) {
		  graph[1+open_atA[i-1].close_index][1+open_atB[j-1].close_index] += CARNAC_CONSERVED_NON_CANONICAL_BONUS;
	       }
#endif     
	    }
	 }
      }
      
      
      /*****************************/ 
      /*    weight of the edges    */
      /*****************************/
      for (i = stemsA.nb_metastems; i > 0; i--) {
	 
	 if(open_atA[i-1].open_for) {
	    
	    for (j = stemsB.nb_metastems; j > 0; j--) {
	       
	       if(open_atB[j-1].open_for) {
		  
		  /*INFO_ "********** I = %i, J = %i\n", i, j _INFO;*/
		  
		  for(x = 1+open_atA[i-1].open_sub; x <= (1+open_atA[i-1].close_sub); x++) {
		     if(close_atA[x-1].close_for) {
			
			for(y = 1+open_atB[j-1].open_sub; y <= (1+open_atB[j-1].close_sub); y++) {
			   if(close_atB[y-1].close_for) {
			      
			      /*	 INFO_ "TRYING (i=%i, j=%i, x=%i, y=%i)\n", i, j, x, y _INFO;*/
			      
			      add_edges(metaA, ostemsA, stemsA, open_atA, close_atA, metaB, ostemsB, stemsB, open_atB, close_atB, i,j,x,y, table, graph, compatibles, &olsA, &olsB, &clsA, &clsB);
			   }
			}
		     }
		  }
	       }
	    }
	 }
      }
   }
   
      
}






void
computeTable(const metastem_list stemsA, const stem_open_index_t *open_atA, const stem_close_index_t *close_atA, const metastem_list stemsB, const stem_open_index_t *open_atB, const stem_close_index_t *close_atB, int **table, int **graph) {
   int w, x, y;


   table[0][0] = 0;
   

   for(x = 0; x <= stemsA.nb_metastems; x++) {
      table[x][0] = 0;
   }
   for(y = 0; y <= stemsB.nb_metastems; y++) {
      table[0][y] = 0;
   }
	 

   

   for(x = 1; x <= stemsA.nb_metastems; x++) {

      for(y = 1; y <= stemsB.nb_metastems; y++) {

	 /*INFO_ "table[%2i][%2i] = MIN(", x, y _INFO;*/

	 table[x][y] = MIN(table[x-1][y], table[x][y-1]);
	 
	 /*INFO_ "%5i, %5i", table[x-1][y], table[x][y-1] _INFO;*/
	 
	 w = graph[x][y];
	 
	 /*INFO_ ", g = %5i", w _INFO;*/
	 
	 if((open_atA[close_atA[x-1].open_index].close_prev >= 0)
	    && (open_atB[close_atB[y-1].open_index].close_prev >= 0)) {

	    /*INFO_ ", g' = %5i", table[1+open_atA[close_atA[x-1].open_index].close_prev][1+open_atB[close_atB[y-1].open_index].close_prev] _INFO;*/

	    w += table[1+open_atA[close_atA[x-1].open_index].close_prev][1+open_atB[close_atB[y-1].open_index].close_prev];
	 }

	 table[x][y] = MIN(table[x][y], w);


	 /*INFO_ ") = %5i\n", table[x][y] _INFO;*/
      }
   }

   /*INFO_ "ENERGY GLOBALE : %i\n", table[stemsA.nb_metastems][stemsB.nb_metastems] _INFO;*/
}








void
build_optimal_folding(const int i, const int k, const metastem_t_list ostemsA, const stem_open_index_t *open_atA, const stem_close_index_t *close_atA, const int j, const int l, const metastem_t_list ostemsB, const stem_open_index_t *open_atB, const stem_close_index_t *close_atB, int **table, stack *p, sankoff_result_pair res) {

   int x, y;
   int m;

   x = k;
   y = l;

   /*INFO_ "  OPTIMAL FOLDING : i = %i, k = %i, j = %i, l = %i\n",i,k,j,l _INFO;*/
   

   while((x>=i) || (y>=j)) {
      if ((y>=j) && (table[x][y] == table[x][y-1])){
	 /*INFO_ "  SKIP x = %i, y = %i (%i) TO x = %i, y = %i (%i)\n", x, y, table[x][y].energy,x,y-1, table[x][y-1].energy _INFO;*/
	 y--; 
      } else if((x>=i) && (table[x][y] == table[x-1][y])) {
	 /*INFO_ "  SKIP x = %i, y = %i (%i) TO x = %i, y = %i (%i)\n", x, y, table[x][y].energy,x-1,y, table[x-1][y].energy _INFO;*/
	 x--;
      } else {

	 /*INFO_ "  TAKE x = %i, y = %i (%i) JUMP TO x = %i, y = %i (%i, stem+under = %i)\n", x, y, table[x][y].energy, 1+open_atA[close_atA[x-1].open_index].close_prev, 1+open_atB[close_atB[y-1].open_index].close_prev, table[1+open_atA[close_atA[x-1].open_index].close_prev][1+open_atB[close_atB[y-1].open_index].close_prev].energy, table[x][y].stem_energy _INFO;*/


	 if(!(open_atA[close_atA[x-1].open_index].is_hairpin || open_atB[close_atB[y-1].open_index].is_hairpin)) {
	    /* should look for subopt */
	    push(p, x, y);
	    /*INFO_ "  PUSH %i, %i\n", x, y _INFO;*/
	 }

	 if(res->nb_stems == 0) {
	    NEW(res->stems1, metastem_t, 1);
	    NEW(res->stems2, metastem_t, 1);
	 } else {
	    RENEW(res->stems1, metastem_t, (res->nb_stems+1));
	    RENEW(res->stems2, metastem_t, (res->nb_stems+1));
	 }

	 res->stems1[res->nb_stems] = ostemsA->metastems[close_atA[x-1].stem_id];
	 NEW(res->stems1[res->nb_stems].seq_stems, stem_t, ostemsA->metastems[close_atA[x-1].stem_id].nb_seqs);
	 for(m = 0; m < ostemsA->metastems[close_atA[x-1].stem_id].nb_seqs; m++) {
	    res->stems1[res->nb_stems].seq_stems[m] = ostemsA->metastems[close_atA[x-1].stem_id].seq_stems[m];
	 }
	 res->energy1 += ostemsA->metastems[close_atA[x-1].stem_id].energy;

	 res->stems2[res->nb_stems] = ostemsB->metastems[close_atB[y-1].stem_id];
	 NEW(res->stems2[res->nb_stems].seq_stems, stem_t, ostemsB->metastems[close_atB[y-1].stem_id].nb_seqs);
	 for(m = 0; m < ostemsB->metastems[close_atB[y-1].stem_id].nb_seqs; m++) {
	    res->stems2[res->nb_stems].seq_stems[m] = ostemsB->metastems[close_atB[y-1].stem_id].seq_stems[m];
	 }
	 res->energy2 += ostemsB->metastems[close_atB[y-1].stem_id].energy;
	 
	 res->nb_stems++;
	 
	 x = 1+open_atA[close_atA[x-1].open_index].close_prev;
	 y = 1+open_atB[close_atB[y-1].open_index].close_prev;
      }
   }
}







void
trace_back(const metastem_t_list ostemsA, const metastem_list stemsA, const stem_open_index_t *open_atA, const stem_close_index_t *close_atA, const metastem_t_list ostemsB, const metastem_list stemsB, const stem_open_index_t *open_atB, const stem_close_index_t *close_atB, int **table, int **graph, sankoff_result_pair res) {
   stack *p = NULL;
   int x, y;


   NEW(p, stack, 1);

   init_stack(p, (stemsA.nb_metastems + stemsB.nb_metastems)*2);
   
   /*INFO_ "TRACE BACK\n" _INFO;*/

   computeTable(stemsA, open_atA, close_atA, stemsB, open_atB, close_atB, table, graph);

   
   x = stemsA.nb_metastems;
   y = stemsB.nb_metastems;

   build_optimal_folding(1, x, ostemsA, open_atA, close_atA, 1, y, ostemsB, open_atB, close_atB, table, p, res);


   while(!is_empty_stack(p)) {
      pop(p, &x,&y);
      /*INFO_ "POP %i, %i\n", x, y _INFO;*/
      
      computeTable_Cell(table, graph, 1+open_atA[close_atA[x-1].open_index].open_next, ostemsA, open_atA, close_atA, 1+open_atB[close_atB[y-1].open_index].open_next, ostemsB, open_atB, close_atB);
      
      build_optimal_folding(1+open_atA[open_atA[close_atA[x-1].open_index].open_next].open_sub, 1+close_atA[x-1].close_prev, ostemsA, open_atA, close_atA, 1+open_atB[open_atB[close_atB[y-1].open_index].open_next].open_sub, 1+close_atB[y-1].close_prev, ostemsB, open_atB, close_atB, table, p, res);

   }


   free_stack(p);
}






void
sankoff_stems(const metasequence metaA, const metastem_t_list ostemsA, const metastem_list stemsA, const stem_open_index_t *open_atA, const stem_close_index_t *close_atA, const metasequence metaB, const metastem_t_list ostemsB, const metastem_list stemsB, const stem_open_index_t *open_atB, const stem_close_index_t *close_atB, cofoldable_t **compatibles, sankoff_result_pair res) {

   int i, j;
   int **table = NULL;
   int **graph = NULL;

   
   NEW(table, int*, (stemsA.nb_metastems+1));
   NEW(graph, int*, (stemsA.nb_metastems+1));
   for(i = 0; i <= stemsA.nb_metastems; i++) {
      NEW(table[i], int, (stemsB.nb_metastems+1));
      NEW(graph[i], int, (stemsB.nb_metastems+1));
      for(j = 0; j <= stemsB.nb_metastems; j++) {
	 table[i][j] = 0;
	 graph[i][j] = 0;
      }	 
   }

   build_alignment_graph(metaA, ostemsA, stemsA, open_atA, close_atA, metaB, ostemsB, stemsB, open_atB, close_atB, table, graph, compatibles);
   
   trace_back(ostemsA, stemsA, open_atA, close_atA, ostemsB, stemsB, open_atB, close_atB, table, graph, res);

   
   for (i = 0; i <= (stemsA.nb_metastems); i++) {
      DESTROY(table[i]);
      DESTROY(graph[i]);
   }
   DESTROY(table);
   DESTROY(graph);

}
