#ifdef HAVE_CONFIG_H
    #include "config.h"
#endif

#include "input.h"
#include "output.h"
#include "opts.h"
#include "display.h"
#include "alignment.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <ctype.h>


int
main(int argc, char **argv) {
   int nb_seqs = 0, i, j;
   sequence* seqs = NULL;
   sequence* seq_t = NULL;
   char *input = NULL;
   int t = 0;
   alignment aln = NULL;

   if(argc != 2) {
      ERROR_ "Missing or too many arguments!!\n" _ERROR;
      return 2;
   }

   init_opts();


   input = argv[1];

   if(read_sequences(input, &seq_t, &t)) {
      ERROR_ "problem while reading %s\n", input _ERROR;
   } else {

      if(nb_seqs) { RENEW(seqs, sequence, (t+nb_seqs)); }
      else { NEW(seqs, sequence, t); }
      memcpy(&(seqs[nb_seqs]), seq_t, t*sizeof(sequence));
      
      nb_seqs += t;
      DESTROY(seq_t);
   }


   aln = init_alignment(seqs, nb_seqs, argv[1]);
   if(aln == NULL) {
      ERROR_ "it's not an alignment !!!\n" _ERROR;
      for(i = 0; i < nb_seqs; i++) {
	 DESTROY(seqs[i]->file);
	 DESTROY(seqs[i]->name);
	 DESTROY(seqs[i]->input_name);
	 DESTROY(seqs[i]->bases);
	 DESTROY(seqs[i]);
      }
      DESTROY(seqs);
      return 2;
   }

   i = 0;
   while(i < aln->length) {
      j = 0;
      while((j < aln->nb_seqs) && !isalpha(aln->seqs[j]->bases[i])) {
	 j++;
      }
      if(j >= aln->nb_seqs) {
	 if(i < (aln->length-1)) {
	    for(j = 0; j < aln->nb_seqs; j++) {
	       memmove(&(aln->seqs[j]->bases[i]), &(aln->seqs[j]->bases[i+1]), sizeof(char)*(aln->length - i - 1));
	    }
	 }
	 for(j = 0; j < aln->nb_seqs; j++) {
	    aln->seqs[j]->length--;
	 }
	 aln->length--;
      } else {
	 i++;
      }
   }

   for(j = 0; j < aln->nb_seqs; j++) {
      aln->seqs[j]->bases[aln->length] = '\0';
   }


   write_clustal(DISPLAY_OUT, aln->seqs, nb_seqs, 1);


   for(i = 0; i < nb_seqs; i++) {
      DESTROY(seqs[i]->file);
      DESTROY(seqs[i]->name);
      DESTROY(seqs[i]->input_name);
      DESTROY(seqs[i]->bases);
      DESTROY(seqs[i]);
   }
   
   DESTROY(seqs);
   
   DESTROY(aln->seqs);
   DESTROY(aln);

   return 0;
}
