#ifdef HAVE_CONFIG_H
    #include "config.h"
#endif
#include "input.h"
#include "output.h"
#include "opts.h"
#include "display.h"


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>


int
main(int argc, char **argv) {
   int nb_seqs = 0, i, w;
   sequence* seqs = NULL, *wseqs = NULL;
   char *output = NULL;
   char *input = NULL;
   FILE* tf = NULL;
   int window_size = 0;
   int overlap = 0;
   int flag, n;


   if(argc != 4) {
      ERROR_ "Missing or too many arguments!!\n" _ERROR;
      return 2;
   }

   init_opts();

   input = argv[1];
   window_size = atoi(argv[2]);
   overlap = atoi(argv[3]);

   if(overlap >= window_size) {
      ERROR_ "Overlp value cannot be greater than window size !!!\n" _ERROR;
      return 3;
   }


   if(read_sequences(input, &seqs, &nb_seqs)) {
      ERROR_ "problem while reading %s\n", input _ERROR;
   } else {
      for(i = 0; i < nb_seqs; i++) {
	 check_sequence(seqs[i]);
      }
   }

   NEW(wseqs, sequence, nb_seqs);
   
   NEW(output, char, (1 + strlen(input) + 20));
   strcpy(output, input);

   flag = 1;
   w = 0;
   n = 1;
   while(flag) {

      flag = 0;

      for(i = 0; i < nb_seqs; i++) {
	 if(seqs[i]->length > w) {
	    wseqs[flag] = init_sequence(seqs[i]->file, seqs[i]->input_name, seqs[i]->name, NULL, 0, seqs[i]->strand);
	    
	    concat_bases(wseqs[flag], &(seqs[i]->bases[w]), MIN(window_size, (seqs[i]->length - w)));
	    
	    flag++;

	 }
      }

      if(flag) {

	 sprintf(&(output[strlen(input)]), ".%i", n);
	 n++;

	 if((tf = fopen(output, "w")) == NULL) {
	    ERROR_ "Unable to open '%s' for writing\n", output _ERROR;
	    return 2;
	 }
	 
	 write_fasta(tf, wseqs, flag, 1);

	 fflush(tf);
	 fclose(tf);

	 w = w + window_size - overlap;
	 
	 for(i = 0; i < nb_seqs; i++) {
	    if(wseqs[i]) {
	       DESTROY(wseqs[i]->file);
	       DESTROY(wseqs[i]->name);
	       DESTROY(wseqs[i]->input_name);
	       DESTROY(wseqs[i]->bases);
	       DESTROY(wseqs[i]);
	    }
	 }
      }
   }


   DESTROY(output);

   for(i = 0; i < nb_seqs; i++) {
      DESTROY(seqs[i]->file);
      DESTROY(seqs[i]->input_name);
      DESTROY(seqs[i]->name);
      DESTROY(seqs[i]->bases);
      
      if(wseqs[i]) {
	 DESTROY(wseqs[i]->file);
	 DESTROY(wseqs[i]->name);
	 DESTROY(wseqs[i]->input_name);
	 DESTROY(wseqs[i]->bases);
	 DESTROY(wseqs[i]);
      }
   }
   
   DESTROY(seqs);
   DESTROY(wseqs);
   
   return 0;
}
